import java.util.*;

/**
 * This class is the main fenhax class 
 * it will recive positions in fen format and generate moves out of these
 *
 * boolean add(String fen)     
 *    Adds given FEN to its fenlist for later move generation
 *    returns true if sucsess, false if FEN is invalid
 *
 * String getMoves(boolean shortNotation, boolean nice)
 *    Returns a String with all moves. If shortNotation is true,
 *    moves is like 1.Ng1-f3, if false 1.Nf3. If nice is true it
 *    adds \t and \n for column mode. Default values is (false,false)
 *
 * String getResult()
 *    If board is in start position but a king or both is in the center
 *    it will determine that result as 1-0, 0-1 or 1/2-1/2
 *
 * void clear()
 *    Clears all
 *                             
 */
public class fenhax {

    static boolean debug=false;
    static boolean rotated=false;

    public fenhax(){
	boards=new ArrayList();
	moves=new movelist();
	diffTable=new ArrayList();
	result="?-?";
	plyCount=0;
    }

    /*
    public fenhax(boolean rotated){
	this.rotated=rotated;
    }
    */

    static String getMoves(boolean shortNotation, boolean nice) { 
	calcMoves();
	return moves.getAll(shortNotation,nice);
    }

    static String getMoves(){
	return getMoves(false,false);
    }

    static int getPlyCount(){
	calcMoves();
	return moves.size();
    }

    static void clear(){
	boards.clear();
	moves.clear();
	diffTable.clear();
	result="?-?";
    }

    static String getResult() { return result; }

    static boolean add(String aFen){
	/*
	if(rotated){
	    String newFen="";
	    for(int i=aFen.length()-1;i>=0;i--){
		newFen += aFen.charAt(i);
	    }
	    aFen=newFen;
	}
	*/

	fenboard f = new fenboard();
	if(!f.setBoard(aFen))
	    return false;

	boards.add(f);

	/** Clear if initial position **/
	if(aFen.indexOf("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR")>=0){
	    clear();
	    boards.add(f);
	}


	/** Check for result **/
	if((aFen.indexOf("rnbq1bnr/pppppppp") >=0 && aFen.indexOf("PPPPPPPP/RNBQ1BNR") >= 0)||
	   (aFen.indexOf("rnbqkbnr/pppppppp") >=0 && aFen.indexOf("PPPPPPPP/RNBQ1BNR") >= 0)||
	   (aFen.indexOf("rnbq1bnr/pppppppp") >=0 && aFen.indexOf("PPPPPPPP/RNBQKBNR") >= 0)){

	    int[] b=f.getBoard();
	    boolean black=false;
	    boolean white=false;

	    for(int i=16;i<=47;i++){
		if(b[i] == 'k') black=true;
		if(b[i] == 'K') white=true;
	    }

	    if(black && white) result="1/2-1/2";
	    else if(black) result="0-1";
	    else if(white) result="1-0";
	}

	if(boards.size()>=2){
	    fenboard last = (fenboard)boards.get(boards.size()-1);
	    fenboard nextlast = (fenboard)boards.get(boards.size()-2);

	    /* Only count changes with less than 5 diffs in position */
	    if(last.diffCount(nextlast) < 5){

		//		System.out.print("!");

		fenboard m=nextlast.mask(last);
		int[] mask = m.getBoard();

   	        for(int i=0;i<64;i++){
		    if(mask[i] != '.' && mask[i]<0){
			diffTable.add(new fendiff(mask[i],i,nextlast,last));
		    }
		}
   	        for(int i=0;i<64;i++){
		    if(mask[i] != '.' && mask[i]>0){
			diffTable.add(new fendiff(mask[i],i,nextlast,last));
		    }
		}

	    }
	    else {
		/** This should be backuped first **/
		clear();
		boards.add(f);

	    }
	}

	return true;
    }

    /** Private functions **/

    static void calcMoves(){
	moves.clear();

	if(debug){
	    for(int i=0;i<diffTable.size();i++){
		System.out.print(i + ":" + ((fendiff)diffTable.get(i)).toString() + " ");
	    }
	    System.out.print("\n");
	}

	int offset=0;
	int length=diffTable.size();

	/* Go through all chunks of possible move. */
	while(offset<length){

	    /** Calculate a chunk of diffs where we may find at least one move by ONE player **/
	    /** Start = offset, end = opponent piece is added OR end of array (length) **/
	    int start=offset;
	    int end=-1;

	    boolean turn = isWhite( ((fendiff)diffTable.get(offset)).getPiece() );

	    int i=start;
	    while(end==-1){
		fendiff test = (fendiff)diffTable.get(i);
		if(turn != isWhite(test.getPiece()) && test.isAdded()){
		    end=i-1;
		}
		else{
		    i++;
		    if(i>=length)
			end=length-1;
		}

		/** Find new move **/
		if(turn==isWhite(test.getPiece()) && test.isAdded()){
		    offset=i;
		}
	    }

	    if(offset==start) offset++;


	    /** Now we have START and END **/
	    if(debug) 	    System.out.print(start + "-"+end+" ");

	    /** Lets find move(s) in chunk. A move is made if a piece is lifted and put
		down on another square **/
	    movelist chunkMoves = new movelist(true,true);

	    for(int s=start;s<end;s++){
		fendiff a = (fendiff)diffTable.get(s);
		for(int e=s+1;e<=end;e++){
		    fendiff b = (fendiff)diffTable.get(e);


		    if(a.getPiece() == b.getPiece() && a.getSquare() != b.getSquare()){
			//if(a.isAdded() && !b.isAdded())
       			//    chunkMoves.add(new move(b,a));
			if(!a.isAdded() && b.isAdded()){
			    String special="";

			    /** Find promotion (if player first moves i.e. e7-e8, then add a queen**/
			    if(b.getPiece()=='P' && b.getSquare()/8==0){
				for(int j=e+1;j<=end;j++){
				    fendiff prom = (fendiff)diffTable.get(j);
				    if(prom.isAdded() && prom.isWhite()){
					if(a.getSquare()%8 != b.getSquare()%8)
					    special+=pos2xy(a.getSquare()).substring(0,1)+"x";
					special+=pos2xy(b.getSquare())+"="+prom.getPiece();
					b=prom;
				    }
				}
			    }
			    if(b.getPiece()=='p' && b.getSquare()/8==7){
				for(int j=e+1;j<=end;j++){
				    fendiff prom = (fendiff)diffTable.get(j);
				    if(prom.isAdded() && prom.isWhite()){
					if(a.getSquare()%8 != b.getSquare()%8)
					    special+=pos2xy(a.getSquare()).substring(0,1)+"x";
					special+=pos2xy(b.getSquare())+"="+prom.getPiece();
					b=prom;
				    }
				}
			    }

			    chunkMoves.add(new move(a,b,special));
			}
		    }
		}
	    }

	    /** Promotion or inserted-piece-move (in Bughouse) **/
	    if(chunkMoves.size()==0){
		fendiff a = (fendiff)diffTable.get(start);
		for(int s=start;s<=end;s++){
		    fendiff b = (fendiff)diffTable.get(s);
		    if(b.isAdded()){
			if(a.getPiece()=='P' && a.getSquare()/8==1 && !a.isAdded()){
			    String special="";
			    if(a.getSquare()%8 != b.getSquare()%8)
				special+=pos2xy(a.getSquare()).substring(0,1)+"x";
			    special+=pos2xy(b.getSquare())+"="+b.getPiece();
			    chunkMoves.add(new move(a,b,special));
			}
			else if(a.getPiece()=='p' && a.getSquare()/8==6 && !a.isAdded()){
			    String special="";
			    if(a.getSquare()%8 != b.getSquare()%8)
				special+=pos2xy(a.getSquare()).substring(0,1)+"x";
			    special+=pos2xy(b.getSquare())+"="+b.getPiece();
			    chunkMoves.add(new move(a,b,special));
			}
			else 
			    chunkMoves.add(new move(a,b,""+a.getPiece()+"@"+pos2xy(a.getSquare())));
		    }
		}
	    }

	    //chunkMoves.removeSmooth();

	    for(int j=0;j<chunkMoves.size();j++){
		if(debug) System.out.print( (chunkMoves.get(j)).toString() + " ");

		moves.add(chunkMoves.get(j));
	    }
	    if(debug)	    System.out.print("\n");
	}
    }

    static String pos2xy(int pos){
	String s = "" + (char)((pos)%8+'a') + "" + (8-pos/8);
	return s;
    }

    static boolean isWhite(char piece){
	if(piece>'A' && piece < 'Z')
	    return true;
	else
	    return false;
    }


    /** Member variables **/
    static movelist moves;
    static ArrayList boards;
    static ArrayList diffTable;
    static String result;
    static int plyCount;

}


class move {
    fendiff from;
    fendiff to;
    boolean take;
    String special;

    public move(fendiff from, fendiff to){
	this.from=from;
	this.to=to;
	take=false;
	special="";
    }

    public move(fendiff from, fendiff to, String special){
	this.from=from;
	this.to=to;
	take=false;
	this.special=special;
    }


    public fendiff getFrom(){ return from; }
    public fendiff getTo(){ return to; }
    public char getPiece(){ return from.getPiece(); }
    public int getFromSquare() { return from.getSquare(); }
    public int getToSquare() { return to.getSquare(); }
    public String getFromCor() { return pos2xy(getFromSquare()); }
    public String getToCor() { return pos2xy(getToSquare()); }
    public boolean isWhite() { return from.isWhite(); }

    public String longNotation() { return moveString(false); }
    public String shortNotation() { return moveString(true); }

    public void setTake(boolean take){ this.take=take; }

    public boolean getTake() { return take; }

    String moveString(boolean shortNotation){
	if(!special.equals("")) {
	    if(isCheck(isWhite()))
		return special + "+";
	    else
		return special;
	}
	       
	boolean ep=false;

	char takeBit='-';
	if(take) takeBit='x';

	String s="";
	if(getPiece()!='p' && getPiece()!='P')
	    s+=new String(""+getPiece()).toUpperCase();

	if(!shortNotation)
	    s+=getFromCor()+takeBit;
	else {

	    /** Check if we need to say where it came from **/
	    if(getPiece()=='R' || getPiece()=='r')
		if(findRookLength(getToSquare()%8,getToSquare()/8,getPiece()))
		    s+=getFromCor()+takeBit;
	    if(getPiece()=='N' || getPiece()=='n')
		if(findNightLength(getToSquare()%8,getToSquare()/8,getPiece()))
		    s+=getFromCor()+takeBit;
	    if(getPiece()=='B' || getPiece()=='b')
		if(findBishopLength(getToSquare()%8,getToSquare()/8,getPiece()))
		    s+=getFromCor()+takeBit;
	    if(getPiece()=='Q' || getPiece()=='q')
		if(findQueenLength(getToSquare()%8,getToSquare()/8,getPiece()))
		    s+=getFromCor()+takeBit;
	    if(getPiece()=='P' || getPiece()=='p')
		if( !getFromCor().substring(0,1).equals( getToCor().substring(0,1) ) ) {
		    s+=getFromCor().substring(0,1)+"x";
		    if(!take) ep=true;
		}

	    if(take && s.length()>=1){
		if(s.charAt(s.length()-1) != 'x')
		    s+="x";
	    }
	}
	
	s+=getToCor();
	if(ep)
	    s+=" ep.";

	/** Is it a check? **/
	if(isCheck(isWhite())) s+='+';

	return s;
    }

    boolean isCheck(boolean white){
	int k=-1;
	int K=-1;
	for(int i=0;i<64;i++){
	    if(to.getAfter().getPieceAt(i%8,i/8) == 'k') k=i;
	    if(to.getAfter().getPieceAt(i%8,i/8) == 'K') K=i;
	}
	boolean check=false;

	if(white){
	    if(findQueenLength(k%8,k/8,'Q')) check=true;
	    if(findRookLength(k%8,k/8,'R')) check=true;
	    if(findBishopLength(k%8,k/8,'B')) check=true;
	    if(findNightLength(k%8,k/8,'N')) check=true;
	    if(findPawnLength(k%8,k/8,'P')) check=true;
	}
	else{
	    if(findQueenLength(K%8,K/8,'q')) check=true;
	    if(findRookLength(K%8,K/8,'r')) check=true;
	    if(findBishopLength(K%8,K/8,'b')) check=true;
	    if(findNightLength(K%8,K/8,'n')) check=true;
	    if(findPawnLength(K%8,K/8,'p')) check=true;
	}

	return check;
    }


    /** Is there a night on a night-distance from square? **/
    boolean findNightLength(int x, int y,char n){
	fenboard b = to.getAfter();

	if(b.getPieceAt(x-1,y-2) == n) return true;
	if(b.getPieceAt(x+1,y-2) == n) return true;
	if(b.getPieceAt(x-2,y-1) == n) return true;
	if(b.getPieceAt(x+2,y-1) == n) return true;
	if(b.getPieceAt(x-2,y+1) == n) return true;
	if(b.getPieceAt(x+2,y+1) == n) return true;
	if(b.getPieceAt(x-1,y+2) == n) return true;
	if(b.getPieceAt(x+1,y+2) == n) return true;

	return false;
    }

    boolean findRookLength(int x, int y, char r){
	fenboard b = to.getAfter();

	int xx=0; 
	int yy=0;

	do{
	    xx++;
	    if(b.getPieceAt(x+xx,y+yy) == r) return true;
	}while(b.getPieceAt(x+xx,y+yy) == '.');

	xx=0;
	do{
	    xx--;
	    if(b.getPieceAt(x+xx,y+yy) == r) return true;
	}while(b.getPieceAt(x+xx,y+yy) == '.');

	xx=0;
	do{
	    yy--;
	    if(b.getPieceAt(x+xx,y+yy) == r) return true;
	}while(b.getPieceAt(x+xx,y+yy) == '.');

	yy=0;
	do{
	    yy++;
	    if(b.getPieceAt(x+xx,y+yy) == r) return true;
	}while(b.getPieceAt(x+xx,y+yy) == '.');

	return false;
    }

    boolean findBishopLength(int x, int y, char r){
	fenboard b = to.getAfter();

	int xx=0; 
	int yy=0;

	do{
	    xx++;
	    yy++;
	    if(b.getPieceAt(x+xx,y+yy) == r) return true;
	}while(b.getPieceAt(x+xx,y+yy) == '.');

	xx=0;
	yy=0;
	do{
	    xx--;
	    yy--;
	    if(b.getPieceAt(x+xx,y+yy) == r) return true;
	}while(b.getPieceAt(x+xx,y+yy) == '.');

	xx=0;
	yy=0;
	do{
	    yy--;
	    xx++;
	    if(b.getPieceAt(x+xx,y+yy) == r) return true;
	}while(b.getPieceAt(x+xx,y+yy) == '.');

	yy=0;
	xx=0;
	do{
	    yy++;
	    xx--;
	    if(b.getPieceAt(x+xx,y+yy) == r) return true;
	}while(b.getPieceAt(x+xx,y+yy) == '.');

	return false;
    }

    boolean findQueenLength(int x, int y, char q){
	if(findRookLength(x,y,q)) return true;
	if(findBishopLength(x,y,q)) return true;

	return false;
    }

    boolean findPawnLength(int x, int y, char p){
	fenboard b = to.getAfter();
	if(p=='P'){
	    if(b.getPieceAt(x-1,y-1)==p) return true;
	    if(b.getPieceAt(x+1,y-1)==p) return true;
	}
	else {
	    if(b.getPieceAt(x-1,y-1)==p) return true;
	    if(b.getPieceAt(x+1,y-1)==p) return true;
	}

	return false;
    }

    public String toString(){ 
	return ""+from.toString() + "-" + to.toString();
    }

    public boolean equals(move other){
	if(!from.equals( other.getFrom() )) return false;
	if(!to.equals( other.getTo() )) return false;
	return true;
    }

      String pos2xy(int pos){
          String s = "" + (char)((pos)%8+'a') + "" + (8-pos/8);
          return s;
      }

}

class movelist {
    ArrayList moves;
    boolean unique;
    boolean noSmooth;

    public void clear(){	moves.clear();    }
    public void remove(int n) { moves.remove(n); }

    public movelist(){ moves=new ArrayList(); unique=false; }
    public movelist(boolean unique, boolean noSmooth) {
	moves=new ArrayList();
	this.unique=unique;
	this.noSmooth=noSmooth;
    }

    public move get(int n){ return (move)moves.get(n); }
    public int size() { return moves.size(); }

    public String toString() { return getAll(false,false); }
    public void print(boolean shortNotation) { 
	System.out.println("Moves");
	System.out.print(getAll(shortNotation,true));
    }

    void fixTakes(){
	for(int i=1;i<moves.size();i++){
	    int toSquare = ((move)moves.get(i)).getToSquare();
	    char p = (char) ((move)moves.get(i-1)).getTo().getAfter().getBoard()[toSquare];
	    if(p!='.'){
		move m=(move)moves.get(i);
		m.setTake(true);
		moves.set(i,m);
	    }
	}


    }


		     
    public String getAll(boolean shortNotation,boolean nice){
	fixTakes();

	       

	String output="";
	boolean lastColor=false;
	int n=1;
	for(int i=0;i<moves.size();i++){
	    move m=(move)moves.get(i);

	    if(m.isWhite() && lastColor==false){
		if(nice)
		    output += n+".\t";
		else
		    output += n+". ";
		n++;
	    }

	    if(n==1 && !m.isWhite()) {
		if(nice)
		    output += "1.\t...\t\t";
		else
		    output += "1. ... ";
	    }

	    lastColor=m.isWhite();
	    int length=output.length();

	    if(shortNotation)
		output += m.shortNotation()+" ";
	    else
		output += m.longNotation()+" ";

	    if(nice){
		if(!m.isWhite())
		    output += "\n";
		else{
		    if(output.length()-length < 8) output += "\t";
		    if(output.length()-length < 12) output += "\t";
		}
	    }
	}
	return output;
    }





    public void add(move aMove){
	if(!unique){
	    moves.add(aMove);
	    return;
	}

	for(int i=0;i<moves.size();i++){
	    if( ((move)moves.get(i)).equals(aMove))
		return;
	}


	
	/** Check for o-o and o-o-o **/
	if(moves.size()>=1){
	    move m=(move)moves.get(moves.size()-1);

	    // O-O
	    if(m.moveString(false).equals("Ke1-g1") && aMove.moveString(false).equals("Rh1-f1")){
		moves.remove(moves.size()-1);
		moves.add(new move(aMove.getFrom(),aMove.getTo(),"O-O"));
		return;
	    }
	    // O-O-O
	    if(m.moveString(false).equals("Ke1-c1") && aMove.moveString(false).equals("Ra1-d1")){
		moves.remove(moves.size()-1);
		moves.add(new move(aMove.getFrom(),aMove.getTo(),"O-O-O"));
		return;
	    }
	    // O-O
	    if(m.moveString(false).equals("Ke8-g8") && aMove.moveString(false).equals("Rh8-f8")){
		moves.remove(moves.size()-1);
		moves.add(new move(aMove.getFrom(),aMove.getTo(),"O-O"));
		return;
	    }
	    // O-O-O
	    if(m.moveString(false).equals("Ke8-c8") && aMove.moveString(false).equals("Ra8-d8")){
		moves.remove(moves.size()-1);
		moves.add(new move(aMove.getFrom(),aMove.getTo(),"O-O-O"));
		return;
	    }
	}

	moves.add(aMove);

    }

    public void removeSmooth(){

	/** We have some serious bugs around here **/

	/** This will remove i.e. first and last of "p&d7-p@d5 p&d7-p@d4 p&d5-p@d4" */
	for(int a=0;a<moves.size();a++){
	    move aMove=(move)moves.get(a);

	    for(int i=0;i<moves.size();i++){
		move test = (move)moves.get(i);
		if( test.getPiece() == aMove.getPiece()){
		    if( aMove.getFrom().getSquare() == test.getTo().getSquare() || 
                        aMove.getTo().getSquare() == test.getFrom().getSquare() ){

			/** BUG?! **/
			if(a==i){
			    moves.remove(i);
			    i=moves.size();
			    a=0;
			}
			else if(i>a){
			    moves.remove(i);
			    moves.remove(a);
			    i=moves.size();
			    a=0;
			}
			else {
			    moves.remove(a);
			    moves.remove(i);
			    i=moves.size();
			    a=0;
			}



		    }
		}
	    }
	}


    }

}
